# -*- coding: utf-8 -*-

"""This module provides classes related to workload vulnerabilities.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from .vulnerability import *
from .vulnerabilityreport import *
