# -*- coding: utf-8 -*-

"""This module is a stub for vulnerability report objects.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject


@dataclass
class VulnerabilityReport(MutableObject):
    pass


__all__ = [
    'VulnerabilityReport',
]
