# -*- coding: utf-8 -*-

"""This module is a stub for classes related to vulnerability exposure scores.

Copyright:
    © 2022 Illumio

License:
    Apache2, see LICENSE for more details.
"""
from dataclasses import dataclass

from illumio.util import MutableObject


@dataclass
class Vulnerability(MutableObject):
    score: int = None


__all__ = [
    'Vulnerability',
]
